﻿using System;
using System.Collections.Generic;
using System.IO;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;

/* Bitmap Font Display Library
 * Created by Brian MacIntosh for Thunderfish Entertainment
 * 
 * Created for: Music Island
 *              7/2011
 * 
 * This class allows you to create a font from a set of images
 * representing different characters.  The images can be stored
 * in two ways:
 * 1.) A 10x10 image containing the characters from ASCII 32 (space)
 *     to ASCII 131 (127 to 131 may be used for special symbol characters).
 * 2.) A set of images named in the form [asciicode].png in their own
 *     folder.  Any byte may be used.  These images must be the same size.
 * 
 * You must call LoadContent(contentManager) after constructing the font object.
 * Strings can then be drawn to the screen using font.DrawString() and variations.
 * 
 * This class also allows animated strings through the BitmapFontAnimation class.
 * Animations are extended from BitmapFontAnimation.  Each BitmapFontAnimation object
 * represents one string that can be drawn anywhere to the screen.  The animation
 * is started by calling animation.Animate(), and must be updated each loop with
 * animation.Update(gameTime).
 * */

namespace Thunderfish
{
    /// <summary>
    /// This class represents a set of images used to display strings.
    /// </summary>
    class BitmapFont
    {
        /// <summary>
        /// The images used by this font.
        /// </summary>
        private Texture2D[] image;

        /// <summary>
        /// Is this font stored in seperate images?
        /// </summary>
        private bool isSplit;

        /// <summary>
        /// The locations of each character within the image.
        /// </summary>
        private Dictionary<byte, Rectangle> grids;

        /// <summary>
        /// The size of each tile in the font image.
        /// </summary>
        private Vector2 gridSize;

        /// <summary>
        /// The file storing the image information for this font.
        /// </summary>
        private string filename;

        /// <summary>
        /// Methods of justifying text.
        /// </summary>
        public enum Justification { LEFT, RIGHT, CENTER }

        /// <summary>
        /// The distance in pixels between each character.
        /// </summary>
        public const int kern = 2;

        /// <summary>
        /// Number of columns on bmfont image
        /// </summary>
        private int tilesAcross;

        /// <summary>
        /// Number of rows on bmfont image
        /// </summary>
        private int tilesDown;

        /// <summary>
        /// The first ascii code that appears in single-image fonts.
        /// </summary>
        private const int ASCII_FIRST = 32;

        /// <summary>
        /// The last ascii code that appears in single-image fonts.
        /// </summary>
        private const int ASCII_LAST = 131;

        /// <summary>
        /// The height of a character in this font.
        /// </summary>
        public int Height { get { return image[0].Height; } }

        /// <summary>
        /// Make a new number font using the images in the given directory.
        /// </summary>
        /// <param name="filename">The path to the image, if the font is stored in one image,
        /// or the folder where the images are stored if seperated.</param>
        /// <param name="isSplit">Are the images in the same file or seperate ones?</param>
        public BitmapFont(string filename, bool isSplit)
        {
            this.filename = filename;
            this.isSplit = isSplit;

            tilesAcross = 10;
            tilesDown = 10;
        }


        /// <summary>
        /// Loads all content associated with this font.
        /// </summary>
        public void LoadContent(ContentManager content)
        {
            if (isSplit)
            {
                string[] files = Directory.GetFiles("Content/" + filename);
                image = new Texture2D[tilesAcross * tilesDown];

                foreach (string str in files)
                {
                    image[ConvertToIndex(Convert.ToInt32(Path.GetFileNameWithoutExtension(str)))]
                        = content.Load<Texture2D>(filename + "/" + Path.GetFileNameWithoutExtension(str));
                }

                //Find size of first non-null image
                int index = 0;
                while (image[index] == null && index < image.Length)
                {
                    index++;
                }
                gridSize = new Vector2(image[index].Width, image[index].Height);
            }
            else
            {
                image = new Texture2D[1];
                image[0] = content.Load<Texture2D>(filename);

                gridSize = new Vector2(image[0].Width / tilesAcross, image[0].Height / tilesDown);
            }

            //Process rectangle dictionary
            grids = new Dictionary<byte, Rectangle>();
            for (byte cha = ASCII_FIRST; cha <= ASCII_LAST; cha++)
            {
                int index = 0;
                if (isSplit) index = ConvertToIndex(cha);

                Rectangle record;
                if (image[index] != null)
                {
                    //Get the tile's data
                    int dataSize = (int)(gridSize.X * gridSize.Y);
                    Color[] imageData = new Color[dataSize];
                    image[index].GetData<Color>(
                        0,
                        GetGrid((char)cha),
                        imageData,
                        0,
                        dataSize);

                    //Find the leftmost non-transparent pixel
                    int left = (int)gridSize.X;
                    for (int c = 0; c < imageData.Length; c++)
                    {
                        if (imageData[c].A > 0)
                        {
                            if (c % (int)gridSize.X < left)
                                left = c % (int)gridSize.X;
                        }
                    }

                    //Find the rightmost non-transparent pixel
                    int right = 0;
                    for (int c = 0; c < imageData.Length; c++)
                    {
                        if (imageData[c].A > 0)
                        {
                            if (c % (int)gridSize.X > right)
                                right = c % (int)gridSize.X;
                        }
                    }

                    //Record
                    record = GetGrid((char)cha);
                    record.X += left;
                    record.Width -= left + ((int)gridSize.X - right);

                    //Special case for "space" character
                    if (cha == 32)
                    {
                        record.X = 16;
                        record.Width = 32;
                    }
                }
                else
                    record = Rectangle.Empty;

                grids.Add(cha, record);
            }
        }


        /// <summary>
        /// Converts an ascii code to an array index.
        /// </summary>
        /// <param name="ascii"></param>
        /// <returns></returns>
        protected int ConvertToIndex(int ascii)
        {
            return ascii - ASCII_FIRST;
        }


        /// <summary>
        /// Converts an array index to the ascii code of the character it stores
        /// </summary>
        /// <param name="index"></param>
        /// <returns></returns>
        protected int ConvertToAscii(int index)
        {
            return index + ASCII_FIRST;
        }


        /// <summary>
        /// Format a number string for display by inserting commas.
        /// </summary>
        /// <param name="text"></param>
        /// <returns></returns>
        public static string FormatNumber(string text)
        {
            //Insert commas
            string ret = "";

            char[] array = text.ToCharArray();
            int reverseDigit = 0;
            for (int c = array.Length - 1; c >= 0; c--)
            {
                reverseDigit++;
                ret = array[c] + ret;
                if (reverseDigit % 3 == 0 && c != 0)
                    ret = "," + ret;
            }

            return ret;
        }


        /// <summary>
        /// Format a time string for display by removing hours and milliseconds.
        /// </summary>
        /// <param name="time"></param>
        /// <returns></returns>
        public static string FormatTime(string time)
        {
            //Infinity exception
            if (time.Length <= 1)
                return time;

            //Remove hours and millis
            time = time.Substring(3, 5);

            return time;
        }


        /// <summary>
        /// Draws the specified string to the screen vertically
        /// </summary>
        /// <param name="batch"></param>
        /// <param name="text"></param>
        /// <param name="position"></param>
        /// <param name="scale"></param>
        /// <param name="align"></param>
        /// <param name="color"></param>
        public void DrawStringVertical(SpriteBatch batch, string text, Vector2 position, float scale, Justification align, Color color)
        {
            int height = 0;
            char[] array = text.ToCharArray();

            //Calculate height of string
            for (int c = 0; c < array.Length; c++)
                height += grids[(byte)array[c]].Height + kern;
            height = (int)(height * scale);

            //Adjust position of string
            switch (align)
            {
                case Justification.CENTER:
                    position -= new Vector2(0, height / 2);
                    break;
                case Justification.LEFT:

                    break;
                case Justification.RIGHT:
                    position -= new Vector2(0, height);
                    break;
            }

            //Draw string
            int tempoffset = 0;
            for (int c = 0; c < array.Length; c++)
            {
                int index = 0;
                if (isSplit) index = ConvertToIndex((byte)array[c]);
                if (image[index] != null)
                {
                    batch.Draw(
                        image[index],
                        position + new Vector2(0, tempoffset),
                        grids[(byte)array[c]],
                        color,
                        0.0f,
                        Vector2.Zero,
                        scale,
                        SpriteEffects.None,
                        0.0f);
                    tempoffset += (int)((grids[(byte)array[c]].Height + kern) * scale);
                }
            }
        }


        /// <summary>
        /// Get the width of the specified string.
        /// </summary>
        /// <param name="str"></param>
        /// <param name="scale"></param>
        public int GetStringWidth(string str, float scale)
        {
            char[] array = str.ToCharArray();
            int width = 0;
            for (int c = 0; c < array.Length; c++)
                width += grids[(byte)array[c]].Width + kern;
            return (int)(width * scale);
        }


        /// <summary>
        /// Draw a string to the screen using this font.
        /// </summary>
        /// <param name="batch"></param>
        /// <param name="text"></param>
        /// <param name="position"></param>
        /// <param name="scale"></param>
        /// <param name="align"></param>
        /// <param name="color"></param>
        public void DrawString(
            SpriteBatch batch,
            string text, Vector2 position, float scale, Justification align, Color color)
        {
            DrawString(batch, new float[text.Length], new Vector2[text.Length],
                text, position, scale, align, color);
        }


        /// <summary>
        /// Draws the specified string to the screen using this font.
        /// </summary>
        public void DrawString(
            SpriteBatch batch,
            float[] charAngles, Vector2[] charOffsets,
            string text, Vector2 position, float scale, Justification align, Color color)
        {
            int width = 0;
            char[] array = text.ToCharArray();

            if (charAngles == null) charAngles = new float[array.Length];
            if (charOffsets == null) charOffsets = new Vector2[array.Length];

            //Calculate width of string
            width = GetStringWidth(text, scale);

            //Adjust position of string
            switch (align)
            {
                case Justification.CENTER:
                    position -= new Vector2(width / 2, 0);
                    break;
                case Justification.LEFT:

                    break;
                case Justification.RIGHT:
                    position -= new Vector2(width, 0);
                    break;
            }

            //Draw string
            int tempoffset = 0;
            for (int c = 0; c < array.Length; c++)
            {
                int index = 0;
                if (isSplit) index = ConvertToIndex((byte)array[c]);
                if (image[index] != null)
                {
                    batch.Draw(
                        image[index],
                        position + new Vector2(tempoffset, 0)
                            + charOffsets[c] + scale * new Vector2(image[index].Width, image[index].Height) / 2,
                        grids[(byte)array[c]],
                        color,
                        charAngles[c],
                        new Vector2(image[index].Width / 2, image[index].Height / 2),
                        scale,
                        SpriteEffects.None,
                        0.0f);
                    tempoffset += (int)((grids[(byte)array[c]].Width + kern) * scale);
                }
            }
        }


        /// <summary>
        /// Draws the specified string to the screen using this font.
        /// </summary>
        /// <param name="destination">The text will fit within this rectangle.</param>
        public void DrawString(SpriteBatch batch,
            string text, Rectangle destination, Justification align, Color color)
        {
            int width = 0;
            int height = 0;
            char[] array = text.ToCharArray();

            //Calculate width of string
            for (int c = 0; c < array.Length; c++)
            {
                width += grids[(byte)array[c]].Width + kern;
                height = grids[(byte)array[c]].Height;
            }
            float scaleX = Math.Min((float)destination.Height / height, (float)destination.Width / width);
            float scaleY = (float)destination.Height / height;

            //Adjust position of string
            switch (align)
            {
                case Justification.CENTER:
                    destination.X = destination.X - width / 2 + destination.Width / 2;
                    break;
                case Justification.LEFT:

                    break;
                case Justification.RIGHT:
                    destination.X = destination.X - width + destination.Width / 2;
                    break;
            }

            //Draw string
            int tempoffset = 0;
            for (int c = 0; c < array.Length; c++)
            {
                int index = 0;
                if (isSplit) index = (byte)array[c] - 32;
                batch.Draw(
                    image[index],
                    new Rectangle(
                        destination.X + tempoffset,
                        destination.Y,
                        (int)(grids[(byte)array[c]].Width * scaleX),
                        (int)(grids[(byte)array[c]].Height * scaleY)),
                    grids[(byte)array[c]],
                    color,
                    0.0f,
                    Vector2.Zero,
                    SpriteEffects.None,
                    0.0f);
                tempoffset += (int)((grids[(byte)array[c]].Width + kern) * scaleX);
            }
        }


        /// <summary>
        /// Returns the location of the specified character
        /// </summary>
        /// <param name="ascii"></param>
        /// <returns></returns>
        private Rectangle GetGrid(char character)
        {
            if (isSplit)
            {
                if (image[ConvertToIndex((byte)character)] != null)
                    return new Rectangle(0, 0, image[ConvertToIndex((byte)character)].Width, image[ConvertToIndex((byte)character)].Height);
                else
                    return Rectangle.Empty;
            }

            return new Rectangle(
                (ConvertToIndex((byte)character) % 10) * (int)gridSize.X,
                (int)Math.Floor(ConvertToIndex((byte)character) / 10f) * (int)gridSize.Y,
                (int)gridSize.X,
                (int)gridSize.Y);
        }
    }
}
