﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

/*
 * See BitmapFont.cs for usage and permissions information.
 */

namespace Thunderfish
{
    /// <summary>
    /// This class represents an animated string using a BitmapFont object for display.
    /// </summary>
    abstract class BitmapFontAnimation
    {
        /// <summary>
        /// Text to display.
        /// </summary>
        public string text;

        /// <summary>
        /// Font to display with.
        /// </summary>
        public BitmapFont font;

        /// <summary>
        /// Total time for one letter to animate.
        /// </summary>
        protected TimeSpan letterAnimTime;

        /// <summary>
        /// Time between letters start animate.
        /// </summary>
        private TimeSpan letterAnimDelay;

        /// <summary>
        /// Time before next letter wiggles.
        /// </summary>
        private TimeSpan nextLetter;

        /// <summary>
        /// The most recent letter wiggled.
        /// </summary>
        protected int currentLetter;

        /// <summary>
        /// Time left for each individual letter to wiggle.
        /// </summary>
        protected TimeSpan[] progresses;

        /// <summary>
        /// Return true if the animation is playing.
        /// </summary>
        public bool IsPlaying
        {
            get
            {
                return (currentLetter < progresses.Length);
            }
        }


        public BitmapFontAnimation(BitmapFont font, string text, TimeSpan letterAnimTime, TimeSpan letterAnimDelay)
        {
            this.text = text;
            this.font = font;
            progresses = new TimeSpan[text.Length];
            for (int c = 0; c < progresses.Length; c++)
                progresses[c] = TimeSpan.Zero;

            this.letterAnimTime = letterAnimTime;
            this.letterAnimDelay = letterAnimDelay;

            currentLetter = progresses.Length;
        }


        public virtual void SetText(string text)
        {
            this.text = text;
            progresses = new TimeSpan[text.Length];
        }


        /// <summary>
        /// Start the animation.
        /// </summary>
        public void Animate()
        {
            currentLetter = 0;
            progresses[currentLetter] = letterAnimTime;
            nextLetter = letterAnimDelay;
        }


        /// <summary>
        /// Update the animation.
        /// </summary>
        /// <param name="gameTime"></param>
        public void Update(GameTime gameTime)
        {
            //Start next letter if necessary
            nextLetter -= gameTime.ElapsedGameTime;
            if (nextLetter <= TimeSpan.Zero)
            {
                nextLetter = letterAnimDelay;
                currentLetter++;
                if (currentLetter < progresses.Length)
                    progresses[currentLetter] = letterAnimTime;
            }

            //Update existing letters
            for (int c = 0; c < progresses.Length; c++)
            {
                progresses[c] -= gameTime.ElapsedGameTime;
                if (progresses[c] <= TimeSpan.Zero) progresses[c] = TimeSpan.Zero;
            }
        }


        public abstract void Draw(SpriteBatch batch, Vector2 position, float scale, BitmapFont.Justification align, Color color);
    }


    /// <summary>
    /// This class represents a text animation that makes the letters wiggle.
    /// </summary>
    class WiggleText : BitmapFontAnimation
    {
        public WiggleText(BitmapFont font, string text, TimeSpan letterAnimTime, TimeSpan letterAnimDelay)
            : base(font, text, letterAnimTime, letterAnimDelay)
        {

        }


        public override void Draw(SpriteBatch batch, Vector2 position, float scale, BitmapFont.Justification align, Color color)
        {
            float[] charAngles = new float[text.Length];
            for (int c = 0; c < charAngles.Length; c++)
                charAngles[c] =
                    (float)Math.Sin(progresses[c].TotalSeconds * 40) / 5f;

            font.DrawString(
                batch,
                charAngles,
                null,
                text,
                position,
                scale,
                align,
                color);
        }
    }


    /// <summary>
    /// This class represents a text animation that makes the letters bounce sequentially.
    /// </summary>
    class BounceText : BitmapFontAnimation
    {
        private int height = 0;

        public BounceText(BitmapFont font, string text, TimeSpan letterAnimTime, TimeSpan letterAnimDelay, int height)
            : base(font, text, letterAnimTime, letterAnimDelay)
        {
            this.height = height;
        }


        public override void Draw(SpriteBatch batch, Vector2 position, float scale, BitmapFont.Justification align, Color color)
        {
            Vector2[] charOffsets = new Vector2[text.Length];
            for (int c = 0; c < charOffsets.Length; c++)
                charOffsets[c] = new Vector2(0, (float)(
                    height * progresses[c].TotalSeconds * (letterAnimTime.TotalSeconds - progresses[c].TotalSeconds)
                    / (letterAnimTime.TotalSeconds * letterAnimTime.TotalSeconds)));

            font.DrawString(
                batch,
                new float[text.Length],
                charOffsets,
                text,
                position,
                scale,
                align,
                color);
        }
    }


    /// <summary>
    /// This class represents an animation that makes the letters change with a slot-machine effect when altered.
    /// </summary>
    class SlotText : BitmapFontAnimation
    {
        public string oldText;


        public SlotText(BitmapFont font, string text, string oldText, TimeSpan letterAnimTime, TimeSpan letterAnimDelay)
            : base(font, text, letterAnimTime, letterAnimDelay)
        {
            this.oldText = oldText;
        }


        public override void SetText(string text)
        {
            base.SetText(text);
            progresses = new TimeSpan[Math.Max(text.Length, oldText.Length)];
        }


        public override void Draw(SpriteBatch batch, Vector2 position, float scale, BitmapFont.Justification align, Color color)
        {
            //Standardize lengths
            int len = Math.Max(oldText.Length, text.Length);
            if (oldText.Length < len) for (int c = 0; c < len - oldText.Length; c++) oldText += " ";
            if (text.Length < len)
            {
                string spaces = "";
                for (int c = 0; c < len - text.Length; c++) spaces += " ";
                SetText(text + spaces);
            }

            Vector2[] charOffsets = new Vector2[text.Length];
            for (int c = 0; c < Math.Min(progresses.Length, charOffsets.Length); c++)
                charOffsets[c] = new Vector2(0, (float)(-1000 * progresses[c].TotalSeconds * (progresses[c].TotalSeconds - letterAnimTime.TotalSeconds)));

            string drawText;
            if (text.Length > currentLetter)
                drawText = text.Substring(0, currentLetter) + oldText.Substring(currentLetter);
            else
                drawText = text;

            font.DrawString(
                batch,
                new float[text.Length],
                charOffsets,
                drawText,
                position,
                scale,
                align,
                color);
        }
    }
}
